/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 3                                                                */
/*                                                                       */
/* Coded by:   L.W. Finger                                               */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            finger@gl.ciw.edu                                          */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/* martin@oc2.oc.chemie.th-darmstadt.de                                  */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

/*
 This file contains routines: build_poly_list, check_color,
          crystal_compute, draw_cell, generate_spheres, import_cif,
          import_schakal
*/

#include "drawxtl.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

extern float dist();

/* global variables for lattice constants and symmetry operators */

extern char spg[3][4];                 /* space group code */
extern int ss[24][3][3];               /* rotation operators for symmetry */
extern float lat_pos[4][3];            /* coordinates of lattice points */
extern float ts[24][3];                /* translation operators */
extern int sym_op_no[192];             /* saved symmetry operation number */
extern float cell_xyz[192][3];         /* storage for generated equiv. positions */
extern int sys;                        /* crystal system indicator */
extern int nbr;                        /* Bravais lattice indicator */
extern int ns;                         /* center of symmetry indicator */
extern int ng;                         /* number of rotation operations to specify group */
extern int nlat;                       /* number of lattice positions */
extern float lat_con[6];               /* lattice constants */
extern float ginv[3][3];               /* direct space metric tensor */
extern float b_inv_mat[3][3];          /* inverse metric tensor */
extern float b_mat[3][3];              /* metric tensor */
extern float rec_lat_con[3];           /* reciprocal lengths */

/* global variables for file names and file variables */

extern char  savedpath[100];           /* path at start of program */
extern char  File_Name[100];           /* Modified Input File Name */
extern FILE  *fpin,*flout;
extern FILE  *fpoutv;
extern FILE  *fpoutp;
extern char  *Str_File_Inp;            /* Input File Name */
extern char  *Str_File;
extern char  *output;

/* global variables for atom lists */

extern int   natom;                    /* number of different types of atoms */
extern int   ncell;                    /* number of atoms in asymmetric unit */
extern int   atom_no[MAX_VERTS];       /* master list atom and symmetry number */
extern float xypos[3*MAX_VERTS];       /* master list fractional coordinates */
extern float *o_vert;                  /* crystal coordinates of saved vertices */
extern float *s_vert;                  /* saved coordinates of vertices */
extern int   *vert_sym_no;             /* saved symmetry operator number */
extern int   *poly_list;               /* storage for polygon corner pointers */
extern int   nvert;                    /* number of vertices in list */
extern int   NvertM;                    /* number of vertices in master atom list */
extern float atom_xyz[MAX_ATOM][3];    /* atomic coordinates */
extern char  atom_l[MAX_ATOM][4];      /* atom labels */
extern int   atom_n[MAX_ATOM];         /* input atom number */
extern int   sv_atom_n[MAX_ATOM];      /* saved input atom number */

/* global variables for spheres and ellipsoids */

extern int   nsphere;                  /* number of different types of spheres */
extern int   n_ellips;                 /* no. of aniso. atoms in asymmetric unit */
extern int   do_ellipsoids;            /* non-zero if ellipsoids to be drawn */
extern int   El_Cutout;                /* True if ellipsoids to have an octant removed */
extern float Sphere_Mult;              /* Multiplier for Sphere Size */
extern float SpMult;                   /* initialization for Sphere_Mult */
extern float Ellipsoid_Prob;           /* Fraction of electron dens enclosed by ellipsoid */
extern float Ellipsoid_Scale;          /* Scaling needed to get probability */
extern float Phong_Value;              /* reflectivity value for Phong highlighting */
extern float Phong_Size;               /* Phong highlighting size */
extern char  sphere_l[MAX_SPHERE][4];  /* sphere label */
extern float sphere_size[MAX_SPHERE];  /* sphere size */
extern char  sphere_col_v[MAX_SPHERE][30];/* sphere colors */
extern char  sphere_col_p[MAX_SPHERE][30];/* sphere colors */
extern char  ellips_l[MAX_EL][4];      /* label for anisotropic atom */
extern int   ellips_n[MAX_EL];         /* atom number */
extern int   ell_type[MAX_EL];         /* form of ellipsoid input */
extern float ellips[MAX_EL][6];        /* thermal ellipsoid coefficients */
extern float ellips_RMS[MAX_EL][3];    /* RMS amplitudes */
extern float ellips_EV[MAX_EL][3][3];  /* Eigenvectors */
extern char  ellips_col_v[MAX_EL][30]; /* ellipsoid colors for VRML */
extern char  ellips_col_p[MAX_EL][30]; /* ellipsoid colors for POV */
extern char  Cutout_color[30];         /* color of cutout section */

/* global variables for polyhedra, planes and bonds */

extern int   npoly;                    /* number of different types of polyhedra */
extern int   nplane;                   /* number of different types of planes */
extern int   nbond;                    /* number of different types of bonds */
extern int   draw_list;                /* number of items in polygon draw list */
extern int   *vertex_list;             /* storage for vertices about a cation */
extern int   numb_list;                /* number in polygon list */
extern int   domolcomp;                /* non-zero if molecule completion requested */
extern float polyfudge;                /* should be close to zero to allow only regular polygonal faces */
extern float Bond_Mult;                /* Multiplier for Bond Size */
extern float BndMult;                  /* Bond multiplier initialization */
extern float mol_d;                    /* intramolecular distance for molecular and polyhedral completion */
extern char  poly_l[MAX_POLY][4];      /* polyhedral center label */
extern float poly_size[MAX_POLY];      /* size of polyhedron */
extern char  poly_col_v[MAX_POLY][30]; /*polyhedron colors for VRML */
extern char  poly_col_p[MAX_POLY][30]; /*polyhedron colors for POV */
extern char  plane_col_v[MAX_PLANE][30];/*plane colors for VRML */
extern char  plane_col_p[MAX_PLANE][30];/*plane colors for POV */
extern char  plane_l[MAX_PLANE][4];    /* plane center label */
extern float plane_size[MAX_PLANE];    /* size of plane coordination */
extern char  bond_l1[MAX_BOND][4];     /* label for one end of bond */
extern char  bond_l2[MAX_BOND][4];     /* label for other end */
extern float bond_size[MAX_BOND];      /* bond diameter */
extern float bond_min[MAX_BOND];       /* minimum bond distance */
extern float bond_max[MAX_BOND];       /* maximum bond distance */
extern char  col_bond_v[MAX_BOND][30]; /* color and properties of bonds for VRML */
extern char  col_bond_p[MAX_BOND][30]; /* color and properties of bonds for POV */
 
/* global variables that control size and orientation of output object */

extern float DepthCue;                 /* scale for Z-dependent edge radius */
extern float X_Boxlim;                 /* X box limit read from command line */
extern float Y_Boxlim;                 /* Y box limit read from command line */
extern float Z_Boxlim;                 /* Z box limit read from command line */
extern float X_Origin;                 /* X origin read from command line */
extern float Y_Origin;                 /* Y origin read from command line */
extern float Z_Origin;                 /* Z origin read from command line */
extern float boxlim[3];                /* half limits of plotting box */
extern float origin[3];                /* position of plotting origin */
extern float G_Rot[3][3];              /* Grand rotation matrix - transforms from Cartesian to picture */
extern float lookat_v1[3],lookat_v2[3];/* "lookat" vectors " */
extern float POV_Max[3];               /* Maximum limits of POV after rotation */
extern float POV_Min[3];               /* Minimum limits of POV after rotation */
extern int   boxflag;                  /* Non-zero if 'bounds' command given */
extern int   packflag;                 /* Non-zero if 'pack' command given */
extern int   docell;                   /* non-zero if unit-cell edges to be drawn */
extern int   Display_axes;             /* non-zero if axial triple to be drawn */
extern int   Color_Warning;            /* Warning flag for non-standard color */
extern float Magnification;            /* Magnification factor for image */
extern float Scale;                    /* Scale of diagram */
extern float offset[3];                /* offset for vector triple */
extern float Text_Size;                /* Size to make axis label text */
extern int   Options;                  /* Place to save command-line options flags */
extern char  col_cell_v[30];           /* color and properties of framebars */
extern char  col_edge_v[30];           /* color and properties of polyhedron edges */
extern char  col_cell_p[30];           /* color and properties of framebars */
extern char  col_edge_p[30];           /* color and properties of polyhedron edges */
extern float rad_cell;                 /* radius of unit cell framebars */
extern int   edges;                    /* draw thin black lines around the edges of polyhedra*/
extern float rad_edge;                 /* radius of cylinders on polyhedron edges */
extern float xrot,yrot,zrot;           /* view rotation angles */
extern float cryst_lim[6];             /* limits of view box in cell coords */
extern int   Unit_Cell;                /* non-zero if unit cell to be drawn */
extern int   no_comment;               /* non-zero to inhibit comment lines in VRML */
extern int   M_cameras;                /* non-zero to inhibit multiple cameras */

#define GEOMedit_list char
#define GEOMobj int

void add_vert();
void add_vert_nc();
void analyze_bonds();
void Blank_Strip();
void build_poly_list();
void build_box_contents();
void check_color();
void crystal_compute();
float determinant();
void do_drawing();
void draw_cell();
void find_all_in_box();
void generate_spheres();
void get_atom_id();
void get_input();
void import_schakal();
void Locate_Triple();
void Output_Spheres();
void symop();
void Token_Strip();
         
/* ************************************************************** */
/* ************************************************************** */

void build_poly_list(numb_list,polyno)

/* routine to take list of 'numb_list' indices in 's_vert'
   and find each plane on surface of convex polyhedron.  It then builds
   a list of pointers to these vertices for drawing program.  The algorithm
   is as follows:
   
   1.  If there are 4 vertices, the polyhedron is a tetrahedron and all
       combinations of 3 vertices constitute a face.

   2.  For more than 4 vertices, some planes will be interior.  These are
       found by choosing three points (indicated by i,j,k) and forming two
       vectors defined by V1 = P(j) - P(i) and V2 = P(k) - P(i).  For each
       additional point (l), a third vector V3 = P(l) - P(i) is formed and
       the volume enclosed by these three vectors V = V3 . V1 x V2 is
       calculated.  If this volume is zero, point l lies in the plane.  If
       the plane is on the surface of the polyhedron, all non-zero volumes
       must have the same sign, which may be positive or negative.
*/

int numb_list;          /* number of candidate vertices */
int polyno;             /* number of polygon - for polygon color */

{
  int i,j,k,l,m;                      /* local loop counters */
  float v1[3],v2[3],v3[3];            /* vectors used in calculation */
  float p[3];                         /* coordinates of point i */    
  float vol,sign_z;
  int ind;                            /* indicator of surface plane */
  int *no;                            /* place to keep vertex id's */
  int nplane;                         /* number in plane */
  int imk,jmk,kmk;
  int ii,jj,ll;
  double phi,cosphi;
  float xm,ym,zm,vx,vy,vz,bm,phi0,vxs,vys,vzs,bms;
  float d1,d2,d3,radius;
  int *ns;

  if ((no = (int *) malloc((unsigned) ((4*numb_list) * sizeof (int)))) == NULL) {
    printf("Unable to get no allocation\n");
    return;
  }
  if ((ns = (int *) malloc((unsigned) ((4*numb_list) * sizeof (int)))) == NULL) {
    printf("Unable to get ns allocation\n");
    free(no);
    return;
  }
  radius = rad_edge * 0.005 * Scale;
  if (DepthCue != 0.) {
/*calculate z coordinate after rotation, and scale edge width accordingly*/
    radius=0.;
    for (j=0;j<=2;j++) radius += G_Rot[j][2]*s_vert[3*no[0]+j];
    radius=rad_edge*0.005*Scale+DepthCue*radius;
/*  radius=rad_edge*0.005*Scale+DepthCue*s_vert[3*no[0]+2];*/
  }
  for (i=0;i<numb_list-2;++i) {
    no[0] = vertex_list[i];                      /* first point */
    for (m=0;m<=2;++m) p[m] = s_vert[3*no[0]+m];
    for (j=i+1;j<numb_list-1;++j) {
      no[1] = vertex_list[j];                  /* second point */
      for (k=0;k<=2;++k)v1[k] = s_vert[3*no[1]+k] - p[k];
      for (k=j+1;k<numb_list;++k) {
	if (numb_list == 4) { /* tetrahedron */
          poly_list[draw_list++] = 3;  /* number of vertices for triangle */
          poly_list[draw_list++] = vertex_list[i]+1;
          poly_list[draw_list++] = vertex_list[j]+1;
          poly_list[draw_list++] = vertex_list[k]+1;
          imk=vertex_list[i];
          jmk=vertex_list[j];
          kmk=vertex_list[k];
          d1 = fabs(s_vert[3*imk]-s_vert[3*jmk]) +
               fabs(s_vert[3*imk+1]-s_vert[3*jmk+1]) +
               fabs(s_vert[3*imk+2]-s_vert[3*jmk+2]);
          d2 = fabs(s_vert[3*imk]-s_vert[3*kmk]) +
               fabs(s_vert[3*imk+1]-s_vert[3*kmk+1]) +
               fabs(s_vert[3*imk+2]-s_vert[3*kmk+2]);
          d3 = fabs(s_vert[3*kmk]-s_vert[3*jmk]) +
               fabs(s_vert[3*kmk+1]-s_vert[3*jmk+1]) +
               fabs(s_vert[3*kmk+2]-s_vert[3*jmk+2]);
          if ((d1 >  0.01) && (d2 > 0.01) && (d3 > 0.01)) {
            fprintf(fpoutp," triangle{< %8.5f, %8.5f, %8.5f>,\n",
              s_vert[3*imk],s_vert[3*imk+1],s_vert[3*imk+2]);
            fprintf(fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>  texture{pigment{color %s  }}\n }\n",
              s_vert[3*jmk],s_vert[3*jmk+1],s_vert[3*jmk+2],
              s_vert[3*kmk],s_vert[3*kmk+1],s_vert[3*kmk+2],poly_col_p[polyno]);
            if (edges == 1)  {
              fprintf(fpoutp," cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                s_vert[3*imk],s_vert[3*imk+1],s_vert[3*imk+2],
                s_vert[3*jmk],s_vert[3*jmk+1],s_vert[3*jmk+2],radius);
              fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_edge_p);
              fprintf(fpoutp," cylinder { <%f,%f,%f>,<%f,%f,%f>,%f\n",
                s_vert[3*jmk],s_vert[3*jmk+1],s_vert[3*jmk+2],
                s_vert[3*kmk],s_vert[3*kmk+1],s_vert[3*kmk+2],radius);
              fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_edge_p);
              fprintf(fpoutp," cylinder { <%f,%f,%f>,<%f,%f,%f>,%f \n",
                s_vert[3*kmk],s_vert[3*kmk+1],s_vert[3*kmk+2],
                s_vert[3*imk],s_vert[3*imk+1],s_vert[3*imk+2],radius);
              fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_edge_p);
            }
          }
	} else {
          no[2] = vertex_list[k];          /* third point */
          nplane = 3;
          for (l=0;l<=2;++l)v2[l] = s_vert[3*no[2]+l] - p[l];            /* get v2 */
          sign_z = 0.0;
	  ind = 0;
	  v3[0] = v1[1] * v2[2] - v1[2] * v2[1];
	  v3[1] = v1[2] * v2[0] - v1[0] * v2[2];
	  v3[2] = v1[0] * v2[1] - v1[1] * v2[0];
	  for (l=0;l<numb_list;++l) {
	    vol = 0.0;
	    for (m=0;m<=2;++m) vol = vol + 
	      (s_vert[3*vertex_list[l]+m] - p[m]) * v3[m];
	    if (fabs(vol) < polyfudge) {
 	      if (l > k ) { 
		no[nplane++] = vertex_list[l]; /* add to plane list */
  /*              fprintf(flout,"Warning - new vertex added to plane list\n");*/
	      } else {
		if (( l != i) && (l != j) && (l != k)) {
	          ind = 1;  /* we already have this plane */
	          l = numb_list; /* skip rest of loop */
		}
	      }
	    } else {
/* point is not in plane - check if first out-of-plane point found */
	      if (sign_z == 0.0) {
		sign_z = vol;   /* yes - set sign */
	      } else {
/* if here, not first out of plane - check that vol and sign_z have same sign. 
   If not, set ind=1 to indicate not a plane */
	        if (vol*sign_z < 0.0) {
		  ind = 1;
		  l = numb_list; /* skip rest of inner loop */
		}
	      }
	    }
	  } /* end of loop on l */
	  if (ind == 0) { /* 0 means a surface plane */
	    poly_list[draw_list++] = nplane;  /* number of vertices in plane */
	    for (l=0;l<nplane;++l)poly_list[draw_list++] = no[l]+1;

            fprintf(fpoutp,"/* %d-sided polygon */\n",nplane); 
/* Make vertices consecutive - find center of polygon */
            xm=ym=zm=0.;
            for (ii=0;ii<nplane;ii++){
              xm=xm+s_vert[3*no[ii]];
              ym=ym+s_vert[3*no[ii]+1];
              zm=zm+s_vert[3*no[ii]+2];
            }

            xm=xm/ (float) nplane;
            ym=ym/ (float) nplane;
            zm=zm/ (float) nplane;

/* sort vertices in consecutive order based on minimal angles between vectors */

            ns[0] = no[0];
            for (ii=0;ii<nplane-1;ii++){
              phi0=1000.;
              vxs= s_vert[3*ns[ii]]-xm;
              vys= s_vert[3*ns[ii]+1]-ym;
              vzs= s_vert[3*ns[ii]+2]-zm;
              bms= sqrt( vxs*vxs+vys*vys+vzs*vzs);

              for (jj=0;jj<nplane;jj++) {
                for (ll=0;ll<ii;ll++)
                  if (ns[ll] == no[jj] ) jj++;

                if (ns[ii] != no[jj] && jj <nplane) {
                  vx= s_vert[3*no[jj]]-xm;
                  vy= s_vert[3*no[jj]+1]-ym;
                  vz= s_vert[3*no[jj]+2]-zm;
                  bm= sqrt( vx*vx+vy*vy+vz*vz);
                  cosphi=( vxs* vx + vys*vy+vzs*vz) /(bms*bm);
                  if (cosphi < -1. ) cosphi = -1.;
                  if (cosphi > 1. ) cosphi=1.;
                  phi = acos(cosphi)*180/3.1415926;
                  if (phi <phi0){
                    phi0=phi;
                    ns[ii+1]=no[jj];
                  }
                }
              } 
            } 
            if (numb_list!=4){
              draw_list=draw_list-nplane;
	      for (l=0;l<nplane;++l)
	        poly_list[draw_list++] = ns[l]+1;
            }
            for (l=0;l<nplane-2;++l){
              d1 = fabs(s_vert[3*ns[0]]-s_vert[3*ns[l+1]]) +
                 fabs(s_vert[3*ns[0]+1]-s_vert[3*ns[l+1]+1]) +
                 fabs(s_vert[3*ns[0]+2]-s_vert[3*ns[l+1]+2]);
              d2 = fabs(s_vert[3*ns[0]]-s_vert[3*ns[l+2]]) +
                 fabs(s_vert[3*ns[0]+1]-s_vert[3*ns[l+2]+1]) +
                 fabs(s_vert[3*ns[0]+2]-s_vert[3*ns[l+2]+2]);
              d3 = fabs(s_vert[3*ns[l+2]]-s_vert[3*ns[l+1]]) +
                 fabs(s_vert[3*ns[l+2]+1]-s_vert[3*ns[l+1]+1]) +
                 fabs(s_vert[3*ns[l+2]+2]-s_vert[3*ns[l+1]+2]);
              if ((d1 >  0.01) && (d2 > 0.01) && (d3 > 0.01)) {
                fprintf(fpoutp," triangle {< %8.5f, %8.5f, %8.5f>,\n"
                  ,s_vert[3*ns[0]],s_vert[3*ns[0]+1],s_vert[3*ns[0]+2]);
                fprintf(fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                  s_vert[3*ns[l+1]],s_vert[3*ns[l+1]+1],s_vert[3*ns[l+1]+2],
                  s_vert[3*ns[l+2]],s_vert[3*ns[l+2]+1],s_vert[3*ns[l+2]+2]);
                fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n"
                    ,poly_col_p[polyno]);
              }
            }
            if (edges == 1)  {
              for (l=0;l<nplane-1;++l){
                d1 = fabs(s_vert[3*ns[l]]-s_vert[3*ns[l+1]]) +
                 fabs(s_vert[3*ns[l]+1]-s_vert[3*ns[l+1]+1]) +
                 fabs(s_vert[3*ns[l]+2]-s_vert[3*ns[l+1]+2]);
                if (d1 >  0.01) {
                  fprintf(fpoutp," cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                    s_vert[3*ns[l]],s_vert[3*ns[l]+1],s_vert[3*ns[l]+2],
                    s_vert[3*ns[l+1]],s_vert[3*ns[l+1]+1],s_vert[3*ns[l+1]+2],radius);
                  fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_edge_p);
                }
              } 
              d1 = fabs(s_vert[3*ns[0]]-s_vert[3*ns[nplane-1]]) +
                 fabs(s_vert[3*ns[0]+1]-s_vert[3*ns[nplane-1]+1]) +
                 fabs(s_vert[3*ns[0]+2]-s_vert[3*ns[nplane-1]+2]);
              if (d1 >  0.01) {
                fprintf(fpoutp," cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                  s_vert[3*ns[0]],s_vert[3*ns[0]+1],s_vert[3*ns[0]+2],
                  s_vert[3*ns[nplane-1]],s_vert[3*ns[nplane-1]+1],s_vert[3*ns[nplane-1]+2],radius);
                fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_edge_p);
              }
            }
          }
	} /* end of numb_list != 4 */
      }   /* end of for (k.... */
    }
  } 
  free(no);
  free(ns);
  return;
} /* end of build_poly_list */
/* ************************************************************** */
/* ************************************************************** */

void check_color(input)

char *input;
{


/*
   VRML routine to check color names and replace them with RGB triples
   derived from 'colors.inc' of POV
*/

  if (strncmp(input,"Gray",4) == 0) {
    strncpy(input,".752941 .752941 .752941\0",24);
    return;
  }                            
  if (strncmp(input,"Grey",4) == 0) {
    strncpy(input,".752941 .752941 .752941\0",24);
    return;
  }
  if (strncmp(input,"LightGray",9) == 0) {
    strncpy(input,".658824 .658824 .658824\0",24);
    return;
  }
  if (strncmp(input,"LightGrey",9) == 0) {
    strncpy(input,".658824 .658824 .658824\0",24);
    return;
  }
  if (strncmp(input,"VLightGrey",10) == 0) {
    strncpy(input,".8 .8 .8\0",9);
    return;
  }

  if (strncmp(input,"White",5) == 0) {
    strncpy(input,"1 1 1\0",6);
    return;
  }
  if (strncmp(input,"Red",3) == 0) {
    strncpy(input,"1 0 0\0",6);
    return;
  }
  if (strncmp(input,"Green",5) == 0) {
    strncpy(input,"0 1 0\0",6);
    return;
  }
  if (strncmp(input,"Blue",3) == 0) {
    strncpy(input,"0 0 1\0",6);
    return;
  }
  if (strncmp(input,"Yellow",6) == 0) {
    strncpy(input,"1 1 0\0",6);
    return;
  }
  if (strncmp(input,"Cyan",3) == 0) {
    strncpy(input,"0 1 1\0",6);
    return;
  }
  if (strncmp(input,"Magenta",7) == 0) {
    strncpy(input,"1 0 1\0",6);
    return;
  }
  if (strncmp(input,"Violet",6) == 0) {
    strncpy(input,".309804 .184314 .309804\0",24);
    return;
  }
  if (strncmp(input,"Black",5) == 0) {
    strncpy(input,"0 0 0\0",6);
    return;
  }
  if (strncmp(input,"Orange",6) == 0) {
    strncpy(input,"1 .5 .0\0",24);
    return;
  }
  if (strncmp(input,"Brown",5) == 0) {
    strncpy(input,".647059 .164706 .164706\0",24);
    return;
  }
  if (strncmp(input,"Gray05",6) == 0) {
    strncpy(input,".05 .05 .05\0",12);
    return;
  }
  if (strncmp(input,"Gray10",6) == 0) {
    strncpy(input,".10 .10 .10\0",12);
    return;
  }
  if (strncmp(input,"Gray15",6) == 0) {
    strncpy(input,".15 .15 .15\0",12);
    return;
  }
  if (strncmp(input,"Gray20",6) == 0) {
    strncpy(input,".20 .20 .20\0",12);
    return;
  }
  if (strncmp(input,"Gray25",6) == 0) {
    strncpy(input,".25 .25 .25\0",12);
    return;
  }
  if (strncmp(input,"Gray30",6) == 0) {
    strncpy(input,".30 .30 .30\0",12);
    return;
  }
  if (strncmp(input,"Gray35",6) == 0) {
    strncpy(input,".35 .35 .35\0",12);
    return;
  }
  if (strncmp(input,"Gray40",6) == 0) {
    strncpy(input,".40 .40 .40\0",12);
    return;
  }
  if (strncmp(input,"Gray45",6) == 0) {
    strncpy(input,".45 .45 .45\0",12);
    return;
  }
  if (strncmp(input,"Gray50",6) == 0) {
    strncpy(input,".50 .50 .50\0",12);
    return;
  }
  if (strncmp(input,"Gray55",6) == 0) {
    strncpy(input,".55 .55 .55\0",12);
    return;
  }
  if (strncmp(input,"Gray60",6) == 0) {
    strncpy(input,".60 .60 .60\0",12);
    return;
  }
  if (strncmp(input,"Gray65",6) == 0) {
    strncpy(input,".65 .65 .65\0",12);
    return;
  }
  if (strncmp(input,"Gray70",6) == 0) {
    strncpy(input,".70 .70 .70\0",12);
    return;
  }
  if (strncmp(input,"Gray75",6) == 0) {
    strncpy(input,".75 .75 .75\0",12);
    return;
  }
  if (strncmp(input,"Gray80",6) == 0) {
    strncpy(input,".80 .80 .80\0",12);
    return;
  }
  if (strncmp(input,"Gray85",6) == 0) {
    strncpy(input,".85 .85 .85\0",12);
    return;
  }
  if (strncmp(input,"Gray90",6) == 0) {
    strncpy(input,".90 .90 .90\0",12);
    return;
  }
  if (strncmp(input,"Gray95",6) == 0) {
    strncpy(input,".95 .95 .95\0",12);
    return;
  }
  if (strncmp(input,"DimGray",7) == 0) {
    strncpy(input,".329412 .329412 .329412\0",24);
    return;
  }
  if (strncmp(input,"DimGrey",7) == 0) {
    strncpy(input,".329412 .329412 .329412\0",24);
    return;
  }
  if (strncmp(input,"Aquamarine",10) == 0) {
    strncpy(input,".439216 .858824 .576471\0",24);
    return;
  }
  if (strncmp(input,"BlueViolet",10) == 0) {
    strncpy(input,".62352 .372549 .623529\0",24);
    return;
  }
  if (strncmp(input,"CadetBlue",9) == 0) {
    strncpy(input,".372549 .623529 .623529\0",24);
    return;
  }
  if (strncmp(input,"Coral",5) == 0) {
    strncpy(input,"1.0 .498039 .0\0",24);
    return;
  }
  if (strncmp(input,"CornflowerBlue",14) == 0) {
    strncpy(input,".258824 .258824 .435294\0",24);
    return;
  }
  if (strncmp(input,"DarkGreen",9) == 0) {
    strncpy(input,".184314 .309804 .184314\0",24);
    return;
  }
  if (strncmp(input,"DarkOliveGreen",14) == 0) {
    strncpy(input,".309804 .309804 .184314\0",24);
    return;
  }
  if (strncmp(input,"DarkOrchid",10) == 0) {
    strncpy(input,".6 .196078 .8\0",24);
    return;
  }
  if (strncmp(input,"DarkSlateBlue",13) == 0) {
    strncpy(input,".419608 .137255 .556863\0",24);
    return;
  }
  if (strncmp(input,"DarkSlateGray",13) == 0) {
    strncpy(input,".184314 .309804 .309804\0",24);
    return;
  }
  if (strncmp(input,"DarkSlateGrey",13) == 0) {
    strncpy(input,".184314 .309804 .309804\0",24);
    return;
  }
  if (strncmp(input,"DarkTurquoise",13) == 0) {
    strncpy(input,".439216 .576471 .858824\0",24);
    return;
  }
  if (strncmp(input,"Firebrick",9) == 0) {
    strncpy(input,".556863 .137255 .137255\0",24);
    return;
  }
  if (strncmp(input,"ForestGreen",11) == 0) {
    strncpy(input,".137255 .556863 .137255\0",24);
    return;
  }
  if (strncmp(input,"Gold",4) == 0) {
    strncpy(input,".8 .498039 .196078\0",24);
    return;
  }
  if (strncmp(input,"Goldenrod",9) == 0) {
    strncpy(input,".858824 .858824 .439216\0",24);
    return;
  }
  if (strncmp(input,"GreenYellow",11) == 0) {
    strncpy(input,".576471 .858824 .439216\0",24);
    return;
  }
  if (strncmp(input,"IndianRed",9) == 0) {
    strncpy(input,".309804 .184314 .184314\0",24);
    return;
  }
  if (strncmp(input,"Khaki",5) == 0) {
    strncpy(input,".623529 .623529 .372549\0",24);
    return;
  }
  if (strncmp(input,"LightBlue",9) == 0) {
    strncpy(input,".74902 .847059 .847059\0",24);
    return;
  }
  if (strncmp(input,"LightSteelBlue",14) == 0) {
    strncpy(input,".560784 .560784 .737255\0",24);
    return;
  }
  if (strncmp(input,"LimeGreen",9) == 0) {
    strncpy(input,".196078 .8 .196078\0",24);
    return;
  }
  if (strncmp(input,"Maroon",6) == 0) {
    strncpy(input,".556863 .137255 .419608\0",24);
    return;
  }
  if (strncmp(input,"MediumAquamarine",16) == 0) {
    strncpy(input,".196078 .8 .6\0",24);
    return;
  }
  if (strncmp(input,"MediumBlue",10) == 0) {
    strncpy(input,".196078 .196078 .8\0",24);
    return;
  }
  if (strncmp(input,"MediumForestGreen",17) == 0) {
    strncpy(input,".419608 .556863 .137255\0",24);
    return;
  }
  if (strncmp(input,"MediumGoldenrod",15) == 0) {
    strncpy(input,".917647 .917647 .678431\0",24);
    return;
  }
  if (strncmp(input,"MediumOrchid",12) == 0) {
    strncpy(input,".576471 .439216 .858824\0",24);
    return;
  }
  if (strncmp(input,"MediumSeaGreen",14) == 0) {
    strncpy(input,".258824 .435294 .258824\0",24);
    return;
  }
  if (strncmp(input,"MediumSlateBlue",15) == 0) {
    strncpy(input,".498039 1.0\0",24);
    return;
  }
  if (strncmp(input,"MediumSpringGreen",17) == 0) {
    strncpy(input,".498039 1.0 0\0",24);
    return;
  }
  if (strncmp(input,"MediumTurquoise",15) == 0) {
    strncpy(input,".439216 .858824 .858824\0",24);
    return;
  }
  if (strncmp(input,"MediumVioletRed",15) == 0) {
    strncpy(input,".858824 .439216 .576471\0",24);
    return;
  }
  if (strncmp(input,"MidnightBlue",12) == 0) {
    strncpy(input,".184314 .184314 .309804\0",24);
    return;
  }
  if (strncmp(input,"Navy",4) == 0) {
    strncpy(input,".137255 .137255 .556863\0",24);
    return;
  }
  if (strncmp(input,"NavyBlue",8) == 0) {
    strncpy(input,".137255 .137255 .556863\0",24);
    return;
  }
  if (strncmp(input,"OrangeRed",9) == 0) {
    strncpy(input,"1.0 .498039\0",24);
    return;
  }
  if (strncmp(input,"Orchid",6) == 0) {
    strncpy(input,".858824 .439216 .858824\0",24);
    return;
  }
  if (strncmp(input,"PaleGreen",9) == 0) {
    strncpy(input,".560784 .737255 .560784\0",24);
    return;
  }
  if (strncmp(input,"Pink",4) == 0) {
    strncpy(input,".737255 .560784 .560784\0",24);
    return;
  }
  if (strncmp(input,"Plum",4) == 0) {
    strncpy(input,".917647 .678431 .917647\0",24);
    return;
  }
  if (strncmp(input,"Salmon",6) == 0) {
    strncpy(input,".435294 .258824 .258824\0",24);
    return;
  }
  if (strncmp(input,"SeaGreen",8) == 0) {
    strncpy(input,".137255 .556863 .419608\0",24);
    return;
  }
  if (strncmp(input,"Sienna",6) == 0) {
    strncpy(input,".556863 .419608 .137255\0",24);
    return;
  }
  if (strncmp(input,"SkyBlue",7) == 0) {
    strncpy(input,".196078 .6 .8\0",14);
    return;
  }
  if (strncmp(input,"SlateBlue",9) == 0) {
    strncpy(input,"0 .498039 1\0",12);
    return;
  }
  if (strncmp(input,"SpringGreen",11) == 0) {
    strncpy(input,"0 1 .498039\0",12);
    return;
  }
  if (strncmp(input,"SteelBlue",9) == 0) {
    strncpy(input,".137255 .419608 .556863\0",24);
    return;
  }
  if (strncmp(input,"Tan",3) == 0) {
    strncpy(input,".858824 .576471 .439216\0",24);
    return;
  }
  if (strncmp(input,"Thistle",7) == 0) {
    strncpy(input,".847059 .74902 .847059\0",23);
    return;
  }
  if (strncmp(input,"Turquoise",9) == 0) {
    strncpy(input,".678431 .917647 .917647\0",24);
    return;
  }
  if (strncmp(input,"VioletRed",9) == 0) {
    strncpy(input,".8 .196078 .6\0",14);
    return;
  }
  if (strncmp(input,"Wheat",5) == 0) {
    strncpy(input,".847059 .847059 .74902\0",23);
    return;
  }
  if (strncmp(input,"YellowGreen",11) == 0) {
    strncpy(input,".6 .8 .196078\0",14);
    return;
  }
  if (strncmp(input,"SummerSky",9) == 0) {
    strncpy(input,".22 .69 .87\0",12);
    return;
  }
  if (strncmp(input,"RichBlue",8) == 0) {
    strncpy(input,".35 .35 .67\0",12);
    return;
  }
  if (strncmp(input,"Brass",5) == 0) {
    strncpy(input,".71 .65 .26\0",12);
    return;
  }
  if (strncmp(input,"Copper",6) == 0) {
    strncpy(input,".72 .45 .20\0",12);
    return;
  }
  if (strncmp(input,"Bronze",6) == 0) {
    strncpy(input,".55 .47 .14\0",12);
    return;
  }
  if (strncmp(input,"Bronze2",7) == 0) {
    strncpy(input,".65 .49 .24\0",12);
    return;
  }
  if (strncmp(input,"Silver",6) == 0) {
    strncpy(input,".90 .91 .98\0",12);
    return;
  }
  if (strncmp(input,"BrightGold",10) == 0) {
    strncpy(input,".85 .85 .10\0",12);
    return;
  }
  if (strncmp(input,"OldGold",7) == 0) {
    strncpy(input,".81 .71 .23\0",12);
    return;
  }
  if (strncmp(input,"Feldspar",8) == 0) {
    strncpy(input,".82 .57 .46\0",12);
    return;
  }
  if (strncmp(input,"Quartz",6) == 0) {
    strncpy(input,".85 .85 .95\0",12);
    return;
  }
  if (strncmp(input,"Mica",4) == 0) {
    strncpy(input,"0 0 0\0",6);
    return;
  }
  if (strncmp(input,"NeonPink",8) == 0) {
    strncpy(input,"1 .43 .78\0",10);
    return;
  }
  if (strncmp(input,"DarkPurple",10) == 0) {
    strncpy(input,".53 .12 .47\0",12);
    return;
  }
  if (strncmp(input,"NeonBlue",8) == 0) {
    strncpy(input,".3 .3 1\0",8);
    return;
  }
  if (strncmp(input,"CoolCopper",10) == 0) {
    strncpy(input,".85 .53 .10\0",12);
    return;
  }
  if (strncmp(input,"MandarinOrange",14) == 0) {
    strncpy(input,".89 .47 .20\0",12);
    return;
  }
  if (strncmp(input,"LightWood",9) == 0) {
    strncpy(input,".91 .76 .65\0",12);
    return;
  }
  if (strncmp(input,"MediumWood",10) == 0) {
    strncpy(input,".65 .50 .39\0",12);
    return;
  }
  if (strncmp(input,"DarkWood",8) == 0) {
    strncpy(input,".52 .37 .26\0",12);
    return;
  }
  if (strncmp(input,"SpicyPink",9) == 0) {
    strncpy(input,"1 .11 .68\0",10);
    return;
  }
  if (strncmp(input,"SemiSweetChoc",13) == 0) {
    strncpy(input,".42 .26 .15\0",12);
    return;
  }
  if (strncmp(input,"BakersChoc",10) == 0) {
    strncpy(input,".36 .20 .09\0",12);
    return;
  }
  if (strncmp(input,"Flesh",5) == 0) {
    strncpy(input,".96 .80 .69\0",12);
    return;
  }
  if (strncmp(input,"NewTan",6) == 0) {
    strncpy(input,".92 .78 .62\0",12);
    return;
  }
  if (strncmp(input,"NewMidnightBlue",15) == 0) {
    strncpy(input,"0 0 .61\0",8);
    return;
  }
  if (strncmp(input,"VeryDarkBrown",13) == 0) {
    strncpy(input,".35 .16 .14\0",12);
    return;
  }
  if (strncmp(input,"DarkBrown",9) == 0) {
    strncpy(input,".36 .25 .20\0",12);
    return;
  }
  if (strncmp(input,"DarkTan",7) == 0) {
    strncpy(input,".59 .41 .31\0",12);
    return;
  }
  if (strncmp(input,"GreenCopper",11) == 0) {
    strncpy(input,".32 .49 .46\0",12);
    return;
  }
  if (strncmp(input,"DkGreenCopper",13) == 0) {
    strncpy(input,".29 .46 .43\0",12);
    return;
  }
  if (strncmp(input,"DustyRose",9) == 0) {
    strncpy(input,".52 .39 .39\0",12);
    return;
  }
  if (strncmp(input,"HuntersGreen",12) == 0) {
    strncpy(input,".13 .37 .31\0",12);
    return;
  }
  if (strncmp(input,"Scarlet",7) == 0) {
    strncpy(input,".55 .09 .09\0",12);
    return;
  }
  if (Color_Warning == 0) {
    printf(" **** Warning - Non-standard color \"%s\" encountered. ****\n",input);
    Color_Warning++;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void crystal_compute()

{
  int i;
  char outfil[100];
  float temp;
  float cx,cy,cz,sx,sy,sz;

  if (Str_File == NULL) return;  /* Fail immediately if filename NULL */
/* initialize */
  origin[0] = X_Origin;
  origin[1] = Y_Origin;
  origin[2] = Z_Origin;
  boxlim[0] = X_Boxlim;
  boxlim[1] = Y_Boxlim;
  boxlim[2] = Z_Boxlim;
  docell = Unit_Cell;
  do_ellipsoids = Color_Warning = 0;
  rad_cell=0.02;  
  Sphere_Mult = SpMult;
  Bond_Mult = BndMult;
  Str_File_Inp = Str_File;
  natom = 0;
  nsphere = npoly = nbond = nplane = n_ellips = 1;
  poly_size[0] = 2.5;
  poly_l[0][0] = poly_l[0][1] = poly_l[0][2] = poly_l[0][3] = ' ';
  plane_size[0] = 2.5;
  plane_l[0][0] = plane_l[0][1] = plane_l[0][2] = plane_l[0][3] = ' ';
  sphere_size[0] = 0.0;
  sphere_l[0][0] = sphere_l[0][1] = sphere_l[0][2] = sphere_l[0][3] = ' ';
  POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.;
  POV_Min[0] = POV_Min[1] = POV_Min[2] =  99999.;

  if ((fpin=fopen(Str_File,"r")) == NULL) {
    printf("Cannot Open Input File");
    return;
  }
  strcpy(outfil,Str_File);
  for (i=1;i<strlen(outfil);i++){
    if (outfil[i] == '.' || outfil[i] ==' ')
      break;
  }
  outfil[i]='\0';
  strcat(outfil,".out");
  if ((flout=fopen(outfil,"w")) == NULL) {
    printf("Cannot open listing file\n");
    return;
  }
  outfil[i]='\0';
  strcat(outfil,".wrl");
  if ((fpoutv=fopen(outfil,"w")) == NULL) {
    printf("Cannot open VRML output file\n");
    return;
  }
  outfil[i]='\0';
  strcat(outfil,".pov");
  if ((fpoutp=fopen(outfil,"w")) == NULL) {
    printf("Cannot open POV output file\n");
    return;
  }

  fprintf(flout,"\n\n                DRAWxtl\n");
  fprintf(flout,"  Crystal Structure Drawing Program %s\n\n",VERSION);
  fprintf(flout,"  Copyright 1996, 1997 by Larry Finger and Martin Kroeker\n");
  fprintf(flout,"  (all rights reserved).\n\n");
  fprintf(flout,    "     Combined POVxtl and VRMLxtl\n");
  fprintf(flout,"POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
  fprintf(flout,"VRMLxtl - Virtual Reality Modeling Language V1.0\n");
  printf("\n\n                DRAWxtl\n");
  printf("  Crystal Structure Drawing Program %s\n\n",VERSION);
  printf("  Copyright 1996, 1997 by Larry Finger and Martin Kroeker\n");
  printf("  (all rights reserved).\n\n");
  printf("     Combined POVxtl and VRMLxtl\n");
  printf("POVxtl  - Scenes for Persistence of Vision Ray Tracer (POV-Ray) V2.0\n");
  printf("VRMLxtl - Virtual Reality Modeling Language V1.0\n");

  get_input();

  if (boxlim[0] == 100.0 ) {         
    boxlim[0] = (cryst_lim[3] - cryst_lim[0])*lat_con[0];
    boxlim[1] = (cryst_lim[4] - cryst_lim[1])*lat_con[1];
    boxlim[2] = (cryst_lim[5] - cryst_lim[2])*lat_con[2];
  }

/* Calculate Grand rotation matrix */

  cx = cos(xrot/RAD);
  cy = cos(yrot/RAD);
  cz = cos(zrot/RAD);
  sx = sin(xrot/RAD);
  sy = sin(yrot/RAD);
  sz = sin(zrot/RAD);
  G_Rot[0][0] = cy*cz;
  G_Rot[0][1] = sx*sy*cz-cx*sz;
  G_Rot[0][2] = -cx*sy*cz-sx*sz;
  G_Rot[1][0] = cy*sz;
  G_Rot[1][1] = sx*sy*sz+cx*cz;
  G_Rot[1][2] = -cx*sy*sz+sx*cz;
  G_Rot[2][0] = sy;
  G_Rot[2][1] = -sx*cy;
  G_Rot[2][2] = cx*cy;

/*
  temp =  determinant(G_Rot);
  if (fabs(temp - 1.0) > 0.0001) {
    printf("Error in G_Rot, det = %f",temp);
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
    exit(1);
  }
*/

/* Initialize POV output limits */

  POV_Max[0] = POV_Max[1] = POV_Max[2] = -99999.;
  POV_Min[0] = POV_Min[1] = POV_Min[2] =  99999.;

/* generate all possible atoms to get the size of the drawing */

  NvertM = 1;                 /* make master list empty */
  if ((s_vert = (float *) malloc((unsigned) (3 * (long) (MAX_VERTS * sizeof (float))))) == NULL) {
    printf("Unable to get initial s_vert allocation\n");
    return;
  }
  get_atom_id();             /* rework atom id numbers to pointers */
  build_box_contents();      /* complete the master atom list */
  free(s_vert);
  printf("\nMaster atom list contains %5d atoms.\n",NvertM-1);
  fprintf(flout,"\nMaster atom list contains %5d atoms.\n",NvertM-1);
/* allocate space for the various arrays */
  if ((o_vert = (float *) malloc((unsigned) (12 * (NvertM+10) * sizeof (float)))) == NULL) {
    printf("Unable to get o_vert allocation\n");
    return;
  }
  if ((s_vert = (float *) malloc((unsigned) (12 * (NvertM+10) * sizeof (float)))) == NULL) {
    printf("Unable to get s_vert allocation\n");
    return;
  }
  if ((vert_sym_no = (int *) malloc((unsigned) (2 * (NvertM+10) * sizeof (int)))) == NULL) {
    printf("Unable to get vert_sym_no allocation\n");
    return;
  }
  if ((poly_list = (int *) malloc((unsigned) ((4*NvertM+10) * sizeof (int)))) == NULL) {
    printf("Unable to get poly_list allocation\n");
    return;
  }
  if ((vertex_list = (int *) malloc((unsigned) ((4*NvertM+10) * sizeof (int)))) == NULL) {
    printf("Unable to get vertex_list allocation\n");
    return;
  }
  analyze_bonds();      /* print the bond distance table */
  printf("\nThe 'pov' and 'wrl' files contain:\n");
  if (docell != 0) {  /* if unit-cell corners drawn, add them for scaling */
    float vert[3];
    int saved_Nvert;

    saved_Nvert = NvertM;
    for (i=0;i<=2;++i)
      vert[i] = 0.0;
    add_vert(vert,0,1);                 /* place 0,0,0 in list */
    vert[2] = 1.0;add_vert(vert,0,1);   /* place 0,0,1 in list */
    vert[1] = 1.0;add_vert(vert,0,1);   /* place 0,1,1 in list */
    vert[2] = 0.0;add_vert(vert,0,1);   /* place 0,1,0 in list */
    vert[0] = 1.0;add_vert(vert,0,1);   /* place 1,1,0 in list */
    vert[1] = 0.0;add_vert(vert,0,1);   /* place 1,0,0 in list */
    vert[2] = 1.0;add_vert(vert,0,1);   /* place 1,0,1 in list */
    vert[1] = 1.0;add_vert(vert,0,1);   /* place 1,1,1 in list */
    NvertM = saved_Nvert;                /* reset list count */
  }
  temp = POV_Max[0] - POV_Min[0];
  if (POV_Max[1] - POV_Min[1] > temp) temp = POV_Max[1] - POV_Min[1];
  if (POV_Max[2] - POV_Min[2] > temp) temp = POV_Max[2] - POV_Min[2];
  Scale = 15.0*temp/Magnification;
/* find location for vector triple */
  if (Display_axes == 1) Locate_Triple();
  fprintf(fpoutv,"#VRML V1.0 ascii\n\n");
  fprintf(fpoutv,"# Scene file created by VRMLxtl from %s\n",Str_File_Inp);
  fprintf(fpoutv,"# with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f\n",
    boxlim[0],boxlim[1],boxlim[2],origin[0],origin[1],origin[2]); 
  fprintf(fpoutv,"#  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f\n",
    cryst_lim[0],cryst_lim[3],cryst_lim[1],cryst_lim[4],cryst_lim[2],cryst_lim[5],xrot,yrot,zrot); 

/* Output VRML header */

  fprintf(fpoutv,"DEF BackgroundColor Info { string \"1. 1. 1.\" }\n");
  fprintf(fpoutv,"Separator {\n");
  fprintf(fpoutv," Material {\n");
  fprintf(fpoutv,"  ambientColor    0.1 0.1 0.1\n");
  fprintf(fpoutv,"  diffuseColor    1 1 1\n");
  fprintf(fpoutv,"  specularColor   0.9 0.9 0.9\n");
  fprintf(fpoutv,"  shininess       0.8\n");
  fprintf(fpoutv," }\n");
  Text_Size = 0.0005*Scale;

  if (M_cameras == 0) {
    fprintf(fpoutv,"# Default viewpoints\n");
    fprintf(fpoutv,"DEF Cameras Switch {\n");
    fprintf(fpoutv,"whichChild 0\n");

    fprintf(fpoutv,"DEF 0_0_1____o OrthographicCamera {\n");
    fprintf(fpoutv,"  position        0 0 %8.3f\n",Scale);
    fprintf(fpoutv,"  orientation     0. -1. 0 %8.3f\n",sin( (lat_con[4]-90.)/180.*PI) );
/*  FIXME: something along this line is needed for triclinic structures
   fprintf(fpoutv,"  orientation     0.25 -0.75 0 %8.3f\n",sin( (lat_con[4]-90.)/180.*PI) );
   */
    fprintf(fpoutv," height %8.3f\n",Scale/12.0);
    fprintf(fpoutv,"}\n");

    fprintf(fpoutv,"DEF 1_0_0____o OrthographicCamera {\n");
    fprintf(fpoutv,"  position       %8.3f %8.3f 0\n",Scale*sin(lat_con[3]/180.*PI),12.0*temp/Magnification*cos(lat_con[3]/180.*PI));
    fprintf(fpoutv,"  orientation     0 1 0 %8.3f\n",lat_con[3]/180.*PI);
    fprintf(fpoutv," height %8.3f\n",Scale/12.0);
    fprintf(fpoutv,"}\n");

    fprintf(fpoutv,"DEF 0_1_0____o OrthographicCamera {\n");
    fprintf(fpoutv,"  position      %8.3f %8.3f 0\n",-sin((lat_con[5]-90.)/180.*PI)*Scale,Scale);
    fprintf(fpoutv,"  orientation     -1. %8.3f 0. 1.5708\n",-sin((lat_con[5]-90.)/180.*PI));
    fprintf(fpoutv," height %8.3f\n",Scale/12.0);
    fprintf(fpoutv,"}\n");

    fprintf(fpoutv,"DEF 0_0_1____p PerspectiveCamera {\n");
    fprintf(fpoutv,"  position        0 0 %8.3f\n",Scale/7.);
    fprintf(fpoutv,"  orientation     0. -1. 0 %8.3f\n",sin( (lat_con[4]-90.)/180.*PI) );
/*  FIXME: something along this line is needed for triclinic structures
    fprintf(fpoutv,"  orientation     0.25 -0.75 0 %8.3f\n",sin( (lat_con[4]-90.)/180.*PI) );
 */
    fprintf(fpoutv,"  focalDistance %8.3f\n",Scale/7.);
    fprintf(fpoutv,"  heightAngle   0.7853\n");
    fprintf(fpoutv,"}\n");

    fprintf(fpoutv,"DEF 1_0_0____p PerspectiveCamera {\n");
    fprintf(fpoutv,"  position       %8.3f %8.3f 0\n",Scale/7.*sin(lat_con[3]/180.*PI),Scale*cos(lat_con[3]/180.*PI));
    fprintf(fpoutv,"  orientation     0 1 0 %8.3f\n",lat_con[3]/180.*PI);
    fprintf(fpoutv,"  focalDistance %8.3f\n",Scale/7.);
    fprintf(fpoutv,"  heightAngle   0.7853\n");
    fprintf(fpoutv,"}\n");

    fprintf(fpoutv,"DEF 0_1_0____p PerspectiveCamera {\n");
    fprintf(fpoutv,"  position      %8.3f %8.3f 0\n",-sin((lat_con[5]-90.)/180.*PI)*Scale/7.,Scale/7.);
    fprintf(fpoutv,"  orientation     -1. %8.3f 0. 1.5708\n",-sin((lat_con[5]-90.)/180.*PI));
    fprintf(fpoutv,"  focalDistance %8.3f\n",Scale/7.);
    fprintf(fpoutv,"  heightAngle   0.7853\n");
    fprintf(fpoutv,"}\n");

    fprintf(fpoutv,"} # Cameras\n");

    fprintf(fpoutv,"ShapeHints {\n");
    fprintf(fpoutv,"vertexOrdering CLOCKWISE shapeType SOLID\n");
    fprintf(fpoutv,"}\n");

  } else {
    fprintf(fpoutv,"PerspectiveCamera {\n");
    fprintf(fpoutv,"  position        0 0 %8.3f\n",Scale/7.);
    fprintf(fpoutv,"  orientation     0 0 1 0\n");
    fprintf(fpoutv,"  focalDistance %8.3f\n",Scale/7.);
    fprintf(fpoutv,"  heightAngle   0.7853\n");
    fprintf(fpoutv,"}\n");
    fprintf(fpoutv,"DirectionalLight{ direction 0 0 -1 intensity 0.4}\n");
    fprintf(fpoutv,"DirectionalLight{ direction 0 0 1 intensity 0.4}\n");
    fprintf(fpoutv,"DirectionalLight{ direction 0 -1 0 intensity 0.4}\n");
    fprintf(fpoutv,"DirectionalLight{ direction 0 1 0 intensity 0.4}\n");
    fprintf(fpoutv,"DirectionalLight{ direction -1 0 0 intensity 0.4}\n");
    fprintf(fpoutv,"DirectionalLight{ direction 1 0 0 intensity 0.4}\n");
  }

  if (xrot != 0.0) {
    fprintf(fpoutv," Rotation {");
    fprintf(fpoutv," rotation 1 0 0 %f}\n",xrot/RAD); 
  }
  if (yrot != 0.0){
    fprintf(fpoutv," Rotation {");
    fprintf(fpoutv," rotation 0 1 0 %f}\n",yrot/RAD); 
  }
  if (zrot != 0.0) {
    fprintf(fpoutv," Rotation {");
    fprintf(fpoutv," rotation 0 0 1 %f}\n",zrot/RAD);
  }

/* Output POV header */

  fprintf(fpoutp,"/*\n\n    POVxtl-Persistence of Vision Ray Tracer\n");
  fprintf(fpoutp,"    Crystal Structure Drawing Program %s\n\n",VERSION);
  fprintf(fpoutp,"       Copyright 1996, 1997 by Larry Finger and Martin Kroeker\n");
  fprintf(fpoutp,"          (all rights reserved).\n\n");
  fprintf(fpoutp," Scene file created from %s \n",Str_File_Inp);
  fprintf(fpoutp," with command line options: -b %6.2f %6.2f %6.2f -o %5.2f %5.2f %5.2f \n",
    boxlim[0],boxlim[1],boxlim[2],origin[0],origin[1],origin[2]); 
  fprintf(fpoutp,"  -p %5.2f %5.2f %5.2f %5.2f %5.2f %5.2f -v %5.2f %5.2f %5.2f -m %5.2f\n*/\n\n",
    cryst_lim[0],cryst_lim[3],cryst_lim[1],cryst_lim[4],cryst_lim[2],cryst_lim[5],xrot,yrot,zrot
    ,Magnification);

  fprintf(fpoutp,"#include \"colors.inc\"\n");
  fprintf(fpoutp,"#include \"chars.inc\"\n");
  fprintf(fpoutp,"camera{\n");
  Scale *= 0.9;
  fprintf(fpoutp," location< 0 ,0, %f>\n",Scale);
  fprintf(fpoutp," up <0, 0.1, 0>\n");
  fprintf(fpoutp," right <-0.1, 0, 0>\n");
  fprintf(fpoutp," look_at <0,0,0>\n");
  fprintf(fpoutp,"}\n");
  fprintf(fpoutp,"background{color White}\n");
  fprintf(fpoutp,"light_source {< 0, 0, %f>\n",Scale);
  fprintf(fpoutp," color red 4.0 green 4.0 blue 4.0\n}\n");
  fprintf(fpoutp,"object{\n union{\n");
  fprintf(fpoutp,"\n");
  fprintf(fpoutp,"#declare xrot = %5.1f\n",xrot);
  fprintf(fpoutp,"#declare yrot = %5.1f\n",yrot);
  fprintf(fpoutp,"#declare zrot = %5.1f\n",zrot);
  fprintf(fpoutp,"\n");
  
  do_drawing(output);               /* generate all elements of the drawing */
  fprintf(fpoutp," }\n");
  fprintf(fpoutp,"rotate <xrot, yrot, zrot>\n");
  fprintf(fpoutp,"}\n");
  fclose(fpoutp);
  fprintf(fpoutv,"}\n");
  fclose(fpoutv);
  fclose(flout);
}  /* end of crystal_compute */

/* ************************************************************** */
/* ************************************************************** */

void draw_cell(output,docell)

/* procedure to add unit-cell outline to geometry file */

GEOMedit_list *output;
int docell;        /* non-zero if unit-cell to be drawn */
{
  int i;
  float vert[3];
  int l0;
  float radius;
  static char abc[]="cbao";

  nvert = 0;                  /* initialize vertex list */
  radius = rad_cell * 0.01 * Scale;

  for (i=0;i<=2;++i)
    vert[i] = 0.0;
  add_vert_nc(vert);             /* place origin at 0,0,0 in list */
  
  vert[2] = 1.0;add_vert_nc(vert);             /* place 0,0,1 in list */
  vert[1] = 1.0;add_vert_nc(vert);             /* place 0,1,1 in list */
  vert[2] = 0.0;add_vert_nc(vert);             /* place 0,1,0 in list */
  vert[1] = 0.0;add_vert_nc(vert);             /* place 0,0,0 in list */
  vert[0] = 1.0;add_vert_nc(vert);             /* place 1,0,0 in list */
  vert[1] = 1.0;add_vert_nc(vert);             /* place 1,1,0 in list */
  vert[2] = 1.0;add_vert_nc(vert);             /* place 1,1,1 in list */
  vert[1] = 0.0;add_vert_nc(vert);             /* place 1,0,1 in list */
  vert[2] = 0.0;add_vert_nc(vert);             /* place 1,0,0 in list */
  vert[1] = 1.0;add_vert_nc(vert);             /* place 1,1,0 in list */
  vert[0] = 0.0;add_vert_nc(vert);             /* place 0,1,0 in list */
  vert[2] = 1.0;add_vert_nc(vert);             /* place 0,1,1 in list */
  vert[0] = 1.0;add_vert_nc(vert);             /* place 1,1,1 in list */
  vert[1] = 0.0;add_vert_nc(vert);             /* place 1,0,1 in list */
  vert[0] = 0.0;add_vert_nc(vert);             /* place 0,0,1 in list */

  if (docell !=0){
    fprintf(fpoutv," Separator{\n");
    fprintf(fpoutv,"  Material{\n");
    fprintf(fpoutv,"   diffuseColor %s\n",col_cell_v);
    fprintf(fpoutv,"  }\n");
    fprintf(fpoutv,"  Coordinate3{ point[\n");
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[0],s_vert[1],s_vert[2]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[3],s_vert[4],s_vert[5]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[6],s_vert[7],s_vert[8]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[9],s_vert[10],s_vert[11]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[15],s_vert[16],s_vert[17]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[18],s_vert[19],s_vert[20]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[21],s_vert[22],s_vert[23]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f]\n  }\n",s_vert[24],s_vert[25],s_vert[26]);
    fprintf(fpoutv,"  IndexedLineSet{coordIndex[\n");
    fprintf(fpoutv,"   0,1,2,3,0,4,5,6,7,4,-1,5,3,-1,2,6,-1,7,1,-1]\n");
    fprintf(fpoutv,"  }\n");
    fprintf(fpoutv," }\n");

    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
        s_vert[0],s_vert[1],s_vert[2],s_vert[3],s_vert[4],s_vert[5],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[3],s_vert[4],s_vert[5],s_vert[6],s_vert[7],s_vert[8],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[6],s_vert[7],s_vert[8],s_vert[9],s_vert[10],s_vert[11],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>, %f\n",
        s_vert[9],s_vert[10],s_vert[11],s_vert[12],s_vert[13],s_vert[14],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[12],s_vert[13],s_vert[14],s_vert[15],s_vert[16],s_vert[17],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[15],s_vert[16],s_vert[17],s_vert[18],s_vert[19],s_vert[20],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[18],s_vert[19],s_vert[20],s_vert[21],s_vert[22],s_vert[23],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[21],s_vert[22],s_vert[23],s_vert[24],s_vert[25],s_vert[26],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[24],s_vert[25],s_vert[26],s_vert[27],s_vert[28],s_vert[29],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[30],s_vert[31],s_vert[32],s_vert[33],s_vert[34],s_vert[35],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[36],s_vert[37],s_vert[38],s_vert[39],s_vert[40],s_vert[41],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f\n",
        s_vert[24],s_vert[25],s_vert[26],s_vert[3],s_vert[4],s_vert[5],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
/* LABELS */
    if (Display_axes == 0) {
      float temp[2];
      nvert = 0;                /* initialize vertex list */
      for (i=0;i<=2;++i)
        vert[i] = 0.0;
      vert[2] = 1.0 + 0.5/lat_con[2]; add_vert_nc(vert);
      vert[2] = 0.0;vert[1] = 1.0 + 0.5/lat_con[1];add_vert_nc(vert); 
      vert[1] = 0.0;vert[0] = 1.0 + 0.5/lat_con[0];add_vert_nc(vert); 
      for (i=0;i<=2;++i)
        vert[i] = -0.5/lat_con[i];
      add_vert_nc(vert);
      fprintf(fpoutp,"/* Labels */\n");
      fprintf(fpoutv,"# Labels\n");
      temp[0] = -1.5*Text_Size;
      temp[1] = -3.0*Text_Size;
      for (l0=0;l0<4;l0++) {
        fprintf(fpoutp,"  text { ttf \"timrom.ttf\",\"%c\",0.1,0.\n",abc[l0]);
        fprintf(fpoutp,"    scale <%4.2f,%4.2f,%4.2f>\n",
           Text_Size*10.,Text_Size*10.,Text_Size*10.);
        fprintf(fpoutp,"    translate <%8.5f,%8.5f,0>\n",temp[0],temp[1]);
        fprintf(fpoutp,"    rotate <-xrot, -yrot, -zrot>\n");
        fprintf(fpoutp,"    translate <%8.5f,%8.5f,%8.5f> pigment{color Black}\n"
          ,s_vert[3*l0],s_vert[3*l0+1],s_vert[3*l0+2]);
        fprintf(fpoutp,"  }\n");
        fprintf(fpoutv," Separator {\n  Translation {");
        fprintf(fpoutv," translation %8.5f %8.5f %8.5f }\n"
         ,s_vert[3*l0],s_vert[3*l0+1],s_vert[3*l0+2]);
        if (zrot != 0.0) fprintf(fpoutv,"  Rotation{ rotation 0 0 1 %8.5f}\n",-zrot/RAD);
        if (yrot != 0.0) fprintf(fpoutv,"  Rotation{ rotation 0 1 0 %8.5f}\n",-yrot/RAD);
        if (xrot != 0.0) fprintf(fpoutv,"  Rotation{ rotation 1 0 0 %8.5f}\n",-xrot/RAD);
        fprintf(fpoutv,"  Scale{ scaleFactor %4.2f %4.2f %4.2f }\n"
         ,Text_Size,Text_Size,Text_Size);
        fprintf(fpoutv,"  Material{ diffuseColor 0. 0. 0. }\n");
        fprintf(fpoutv,"  AsciiText { string  \"%c\" }\n",abc[l0]);
        fprintf(fpoutv," }\n");
      }
    }
  }

  if (Display_axes == 1){
    float xysep,temp1[3],temp[3];
	int j;
    nvert = 0;                /* initialize vertex list */
    for (i=0;i<=2;++i)
      vert[i] = 0.0;
    add_vert_nc(vert);         /* place origin at 0,0,0 in list */
    vert[2] = 1.0/lat_con[2];add_vert_nc(vert);               /* place 1 A vector along 001 */
    vert[2] = 0.0;vert[1] = 1.0/lat_con[1];add_vert_nc(vert); /* place 1 A vector along 010 */
    vert[1] = 0.0;vert[0] = 1.0/lat_con[0];add_vert_nc(vert); /* place 1 A vector along 100 */
    vert[2] = 1.4/lat_con[2];vert[0]= 0.0; add_vert_nc(vert); /* place 1.4 A vector along 001 */
    vert[2] = 0.0;vert[1] = 1.4/lat_con[1];add_vert_nc(vert); /* place 1.4 A vector along 010 */
    vert[1] = 0.0;vert[0] = 1.4/lat_con[0];add_vert_nc(vert); /* place 1.4 A vector along 100 */
    for (i=6;i>=0;i--) {        /* Translate coordinates */
      for (j=0;j<3;j++) {
        s_vert[3*i+j] += offset[j] - s_vert[j];
      }
    }
/* draw vector triple in POV */
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
        s_vert[0],s_vert[1],s_vert[2],s_vert[3],s_vert[4],s_vert[5],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
        s_vert[0],s_vert[1],s_vert[2],s_vert[6],s_vert[7],s_vert[8],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
    fprintf(fpoutp,"  cylinder{<%f,%f,%f>,<%f,%f,%f>,%f \n",
        s_vert[0],s_vert[1],s_vert[2],s_vert[9],s_vert[10],s_vert[11],radius);
    fprintf(fpoutp,"   texture{pigment{color %s }}\n",col_cell_p);
    fprintf(fpoutp,"   no_shadow\n  }\n");
/* draw vector triple in VRML */
    fprintf(fpoutv," Separator{\n");
    fprintf(fpoutv,"  Material{\n");
    fprintf(fpoutv,"   diffuseColor %s\n",col_cell_v);
    fprintf(fpoutv,"  }\n");
    fprintf(fpoutv,"  Coordinate3{ point[\n");
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[0],s_vert[1],s_vert[2]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[3],s_vert[4],s_vert[5]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f,\n",s_vert[6],s_vert[7],s_vert[8]);
    fprintf(fpoutv,"   %8.5f %8.5f %8.5f]\n  }\n",s_vert[9],s_vert[10]
      ,s_vert[11]);
    fprintf(fpoutv,"  IndexedLineSet{coordIndex[");
    fprintf(fpoutv,"0,1,-1,0,2,-1,0,3,-1]\n");
    fprintf(fpoutv,"  }\n");
    fprintf(fpoutv," }\n");
    for (l0=0;l0<3;l0++) {
      for(j=0;j<2;j++) {
        int k;
        temp[j] = 0.0;
        temp1[j] = 0.0;
        for (k=0;k<3;k++) {
          temp[j] += G_Rot[k][j] * s_vert[3*l0+k+12];
          temp1[j] += G_Rot[k][j] * s_vert[k];
        }
      }
      xysep = (temp[0] - temp1[0]) * (temp[0] - temp1[0])
        + (temp[1] - temp1[1]) * (temp[1] - temp1[1]);
      if (xysep > 0.1) {   /* only draw label if away from axes */
        temp[0] = -1.5*Text_Size;
        temp[1] = -3.0*Text_Size;
        fprintf(fpoutp,"  text { ttf \"timrom.ttf\",\"%c\",0.1,0.\n",abc[l0]);
        fprintf(fpoutp,"    scale <%4.2f,%4.2f,%4.2f>\n",Text_Size*10.,Text_Size*10.,Text_Size*10.);
        fprintf(fpoutp,"    translate <%8.5f,%8.5f,0>\n",temp[0],temp[1]);
        fprintf(fpoutp,"    rotate <-xrot, -yrot, -zrot>\n");
        fprintf(fpoutp,"    translate <%8.5f,%8.5f,%8.5f> pigment{color Black}\n"
          ,s_vert[3*l0+12],s_vert[3*l0+13],s_vert[3*l0+14]);
        fprintf(fpoutp,"  }\n");
        fprintf(fpoutv," Separator {\n  Translation {");
        fprintf(fpoutv," translation %8.5f %8.5f %8.5f }\n"
         ,s_vert[3*l0+12],s_vert[3*l0+13],s_vert[3*l0+14]);
        if (zrot != 0.0) fprintf(fpoutv,"  Rotation{ rotation 0 0 1 %8.5f}\n",-zrot/RAD);
        if (yrot != 0.0) fprintf(fpoutv,"  Rotation{ rotation 0 1 0 %8.5f}\n",-yrot/RAD);
        if (xrot != 0.0) fprintf(fpoutv,"  Rotation{ rotation 1 0 0 %8.5f}\n",-xrot/RAD);
        fprintf(fpoutv," Translation {");
        fprintf(fpoutv," translation 0 %8.5f 0 }\n"
         ,-Text_Size/2);
        fprintf(fpoutv,"  Scale{ scaleFactor %4.2f %4.2f %4.2f }\n"
         ,Text_Size,Text_Size,Text_Size);
        fprintf(fpoutv,"  Material{ diffuseColor 0. 0. 0. }\n");
        fprintf(fpoutv,"  AsciiText { string  \"%c\" justification CENTER }\n"
           ,abc[l0]);
        fprintf(fpoutv," }\n");
      }
    }
  }

  nvert = 0;                  /* empty vertex list */
}   /* end of draw_cell */

/* ************************************************************** */
/* ************************************************************** */

void generate_spheres(output)

GEOMedit_list *output;

/* routine to generate spherical objects and add then to edit list */

{
  extern int natom;
  float *radii;                     /* dynamic array for sphere radii */  
  int Sphere_Count;                 /* Counter for number output */
  int i,j,k;                        /* loop counters */

  Sphere_Count = 0;
  for (i=1;i<nsphere;++i) {    /* loop through spheres */
    nvert = 0;              /* clear the vertex list */
    for (j=0;j<natom;++j)   /* loop through atoms */
      if( atom_n[j] == -i ) find_all_in_box(j); 
    if (nvert > 0)  {    /* if spheres present */
/* allocate storage for radii of spheres */
      radii = (float *) malloc ((unsigned)(nvert*sizeof(float)));
      for (k=0;k<nvert;++k) /* copy radii info to storage*/
        radii[k] = sphere_size[i] * Sphere_Mult;
      Output_Spheres(radii,output,i);
      free(radii);
    }
    Sphere_Count += nvert;
  }
  if (Sphere_Count > 0) {
    printf("%4d spheres.\n",Sphere_Count);
    fprintf(flout,"Generated %4d spheres.\n",Sphere_Count);
  }
}  /* end of generate_spheres */

/* ************************************************************** */
/* ************************************************************** */

void import_cif(input,in_line)

char input[100];
int in_line;
/* routine to extract structural information from a CIF */

{
  FILE *impin;
  char string[100],filename[100],tstring[100];
  int U_pos[6],X_pos[3],Label_pos;
  int i,j;
  
  if (in_line) {
  fprintf(stderr,"Cannot read inline CIF data. Please use the import instruction\n");
#ifdef _MAC_
  printf("Use Apple Q to close this window.\n");
#endif
  exit(0);
  }
  sscanf(input,"%s %s",filename,filename); /* get file name */

  if ((impin=fopen(filename,"r")) == NULL) {
    fprintf(stderr,"Cannot Open CIF Import File, Run aborted.\n");
#ifdef _MAC_
    printf("Use Apple Q to close this window.\n");
#endif
    exit(0);
  }
  while(strncmp(string,"_cell_length_a",14)!=0) {
    if(!fgets(string,100,impin)){; /* search for a-axis length */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  sscanf(string,"%s %f",tstring,&lat_con[0]);
  rewind(impin);
  while(strncmp(string,"_cell_length_b",14)!=0){
    if(!fgets(string,100,impin)){; /* search for b-axis */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  sscanf(string,"%s %f",tstring,&lat_con[1]);
  rewind(impin);
  while(strncmp(string,"_cell_length_c",14)!=0){
    if(!fgets(string,100,impin)){; /* search for c axis */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  sscanf(string,"%s %f",tstring,&lat_con[2]);
  rewind(impin);
  while(strncmp(string,"_cell_angle_al",14)!=0){
    if(!fgets(string,100,impin)){; /* search for alpha */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  rewind(impin);
  sscanf(string,"%s %f",tstring,&lat_con[3]);
  while(strncmp(string,"_cell_angle_be",14)!=0){
    if(!fgets(string,100,impin)){; /* search for beta */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  rewind(impin);
  sscanf(string,"%s %f",tstring,&lat_con[4]);
  while(strncmp(string,"_cell_angle_ga",14)!=0){
    if(!fgets(string,100,impin)){; /* search for gamma */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  rewind(impin);
  j=1;
  sscanf(string,"%s %f",tstring,&lat_con[5]);
  fprintf(flout,"******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n"
     ,lat_con[0],lat_con[1],lat_con[2],lat_con[3],lat_con[4],lat_con[5]);
  while(strncmp(string,"_symmetry_space",15)!=0){
    if(!fgets(string,100,impin)){; /* search for space group */
      j = 0;         /* indicate not found */
      break;         /* no error if not found */
    }
    Blank_Strip(string);
  }
  if (j != 0) {
    tstring[0] = 's';
    tstring[1] = 'p';
    tstring[2] = 'g';
    tstring[3] = 'r';
    tstring[4] = ' ';
    for (i=0;i<60;i++)  /* change the keyword to spaces */
      if (string[i] == ' ') {
        break;
      } else {
        string[i] = ' ';
      }
    Blank_Strip(string);   /* strip off the spaces */
    if (string[0] != '?' ) {
      for (i=0;i<60;i++)
        if(string[i] == (char)39)break;
      for(i++,j=5;i<60;i++) 
        if(string[i] == (char)39)break;
      else
        tstring[j++] = string[i];
      tstring[j] = '\0';
      fprintf(flout,"******      %s\n",tstring);
      for(i=0;i<60;i++)
        if (tstring[i] == 'H')tstring[i] = ' '; /* remove any H from rhomb. spgrps */
      symop(tstring);
    }
  }
  rewind(impin);
/* skip through file to atom_site */
  while(strncmp(string,"_atom_site_",11)!=0){
    if(!fgets(string,100,impin)){ /* search for phase info */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
/* first _atom_site_ line in string need to find position of label, x, y, and z */
  j = 0;
  while(strncmp(string,"_atom_site_",11)==0){
    if (strncmp(string,"_atom_site_label",16)==0)Label_pos = j;
    if (strncmp(string,"_atom_site_fract_x",18)==0)X_pos[0] = j;
    if (strncmp(string,"_atom_site_fract_y",18)==0)X_pos[1] = j;
    if (strncmp(string,"_atom_site_fract_z",18)==0)X_pos[2] = j;
    j++;
    if(!fgets(string,100,impin)){ /* search for phase info */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  for(;;) {
    if (strlen(string) < 5)break;
    if (strncmp(string,"loop_",5)==0)break;
    atom_n[natom] = 0;
    for (i=0;i<4;i++) atom_l[natom][i] = ' ';  /* initialize atom name */
    strcpy(tstring,string);
    if (Label_pos > 0) Token_Strip(tstring,Label_pos);
    j = 0;
    for (i=0;i<10;i++){
      if (tstring[i] >= '0' && tstring[i] <= '9') {
        atom_n[natom] = 10 * atom_n[natom] + (int)tstring[i]-48;
        atom_l[natom][i] = ' ';
      } else {
        atom_l[natom][j++] = tstring[i];
        if (j > 3 || tstring[i] == ' ') break;
      }
    }
    strcpy(tstring,string);
    if (X_pos[0] > 0)Token_Strip(tstring,X_pos[0]);
    sscanf(tstring,"%f",&atom_xyz[natom][0]);
    strcpy(tstring,string);
    if (X_pos[1] > 0)Token_Strip(tstring,X_pos[1]);
    sscanf(tstring,"%f",&atom_xyz[natom][1]);
    strcpy(tstring,string);
    if (X_pos[2] > 0)Token_Strip(tstring,X_pos[2]);
    sscanf(tstring,"%f",&atom_xyz[natom][2]);
    fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",atom_l[natom][0]
      ,atom_l[natom][1],atom_l[natom][2],atom_l[natom][3],atom_n[natom]
      ,atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
    for(i=0;i<100;i++)string[i]=' ';
    if(!fgets(string,100,impin)) {
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
    if (++natom >= MAX_ATOM) {
      fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
  }
  rewind(impin);
/* skip through file to atom_site_aniso */
  while(strncmp(string,"_atom_site_aniso",16)!=0){
    if(!fgets(string,100,impin)){ /* search for anisotropic U's */
      fclose(impin);
      return;
    }
    Blank_Strip(string);
  }
/* first _atom_site_aniso line in string
   need to find position of label, U11, etc */
  j = 0;
  while(strncmp(string,"_atom_site_aniso",16)==0){
    if (strncmp(string,"_atom_site_aniso_label",22)==0)Label_pos = j;
    if (strncmp(string,"_atom_site_aniso_U_11",21)==0)U_pos[0] = j;
    if (strncmp(string,"_atom_site_aniso_U_22",21)==0)U_pos[1] = j;
    if (strncmp(string,"_atom_site_aniso_U_33",21)==0)U_pos[2] = j;
    if (strncmp(string,"_atom_site_aniso_U_12",21)==0)U_pos[3] = j;
    if (strncmp(string,"_atom_site_aniso_U_13",21)==0)U_pos[4] = j;
    if (strncmp(string,"_atom_site_aniso_U_23",21)==0)U_pos[5] = j;
    j++;
    if(!fgets(string,100,impin)){ /* search for phase info */
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  for(;;) {
    if (strlen(string) < 5)break;
    if (strncmp(string,"_",1)==0)break;
    ellips_n[n_ellips] = 0;
    for (i=0;i<4;i++) ellips_l[n_ellips][i] = ' ';  /* initialize ellipsoid name */
    strcpy(tstring,string);
    if (Label_pos > 0) Token_Strip(tstring,Label_pos);
    j=0;
    for (i=0;i<10;i++) {
      if (tstring[i] >= '0' && tstring[i] <= '9'){
        ellips_n[n_ellips] = 10 * ellips_n[n_ellips] + (int)tstring[i]-48;
        ellips_l[n_ellips][i] = ' ';
      } else {
        ellips_l[n_ellips][j++] = tstring[i];
        if (j > 3 || tstring[i] == ' ') break;
      }
    }
    ell_type[n_ellips] = 1; /* Uij */
    do_ellipsoids = 1;
    strcpy(tstring,string);
    if (U_pos[0] > 0)Token_Strip(tstring,U_pos[0]);
    sscanf(tstring,"%f",&ellips[n_ellips][0]);
    strcpy(tstring,string);
    if (U_pos[1] > 0)Token_Strip(tstring,U_pos[1]);
    sscanf(tstring,"%f",&ellips[n_ellips][1]);
    strcpy(tstring,string);
    if (U_pos[2] > 0)Token_Strip(tstring,U_pos[2]);
    sscanf(tstring,"%f",&ellips[n_ellips][2]);
    strcpy(tstring,string);
    if (U_pos[3] > 0)Token_Strip(tstring,U_pos[3]);
    sscanf(tstring,"%f",&ellips[n_ellips][3]);
    strcpy(tstring,string);
    if (U_pos[4] > 0)Token_Strip(tstring,U_pos[4]);
    sscanf(tstring,"%f",&ellips[n_ellips][4]);
    strcpy(tstring,string);
    if (U_pos[5] > 0)Token_Strip(tstring,U_pos[5]);
    sscanf(tstring,"%f",&ellips[n_ellips][5]);
    fprintf(flout,"******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n"
      ,ellips_l[n_ellips][0],ellips_l[n_ellips][1],ellips_l[n_ellips][2]
      ,ellips_l[n_ellips][3],ellips_n[n_ellips],ellips[n_ellips][0]
      ,ellips[n_ellips][1],ellips[n_ellips][2],ellips[n_ellips][3]
      ,ellips[n_ellips][4],ellips[n_ellips][5]);
    if (++n_ellips >= MAX_EL) {
      fprintf(flout,"Too many ellipsoid types specified - increase MAX_EL.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    if(!fgets(string,100,impin)) {
      fprintf(stderr,"Error reading CIF Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    Blank_Strip(string);
  }
  fclose(impin);
}

/* ************************************************************** */
/* ************************************************************** */

void import_schakal(input,in_line)

char input[100];
int in_line; /*flag for in-line schakal data */

/* routine to extract structural information from a SCHAKAL data file */

{
  FILE *impin;
  char string[100];              /* temporary string */
  char filename[100];            /* string for filename */
  int intype;                    /* type of datum read */
  int i,j;                       /* temporary */

  if (in_line) {
    impin=fpin;
  } else {
    sscanf(input,"%s %s",filename,filename); /* get file name */
    if ((impin=fopen(filename,"r")) == NULL) {
      fprintf(stderr,"Cannot Open SCHAKAL Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
  }
  fprintf(flout,"****** Now interpreting SCHAKAL-style data\n");
  while (!feof(impin)) {
    if(!fgets(input,100,impin)) {             /* read a line */
      fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(1);
    }
    intype = 0;
    for(i=0;i<4;i++) {
      if(input[i] == '\0')intype = 3;
    }
    if(strncmp(input,"TITL",4) == 0)intype = 1; /* echo and ignore */
    if(strncmp(input,"titl ",4) == 0)intype = 1; /* echo and ignore */
    if(strncmp(input,"CELL",4) == 0)intype = 2;
    if(strncmp(input,"cell",4) == 0)intype = 2;
    if(strncmp(input,"ATOM",4) == 0)intype = 3; /* atom coordinate card */
    if(strncmp(input,"atom",4) == 0)intype = 3;
    if(strncmp(input,"END",3) == 0)intype = 4; /* trigger exit*/
    if(strncmp(input,"end",3) == 0)intype = 4; /* trigger exit*/


    switch (intype) {
        case 1:
            for(i=0;i<51;i++) if(input[i] == '\n')input[i] = ' ';
            input[50] = '\0';    /* truncate title */
            fprintf(flout,"******      %s\n",input);
            break;      /* ignore title except for listing */
        case 2: for (i=3;i<=5;++i) lat_con[i] = 0.;
            sscanf(input,"%s %f %f %f %f %f %f",string,&lat_con[0],&lat_con[1]
              ,&lat_con[2],&lat_con[3],&lat_con[4],&lat_con[5]);
            fprintf(flout,"******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n"
              ,lat_con[0],lat_con[1],lat_con[2],lat_con[3],lat_con[4],lat_con[5]);
	    break;
        case 3:    /* read an atom line */
            sscanf(input,"%s %c%c%c%c %f %f %f",string,
              &atom_l[natom][0],&atom_l[natom][1]
             ,&atom_l[natom][2],&atom_l[natom][3],
              &atom_xyz[natom][0],&atom_xyz[natom][1],&atom_xyz[natom][2]);
            atom_n[natom] = 0;
            for (j=0;j<4;j++)   /* extract atom number from SCHAKAL name */
              if (atom_l[natom][j] >= '0' && atom_l[natom][j] <= '9') {
                atom_n[natom] = atom_n[natom]*10 + (int)atom_l[natom][j]-48;
                atom_l[natom][j] = ' ';
              }
            if (atom_n[natom] == 0) atom_n[natom] = 1;
            fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n"
              ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
              ,atom_n[natom],atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
            natom++;
            if (natom >= MAX_ATOM) {
              fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
            break;    
        case 4:
            fprintf(flout,"****** END - end of SCHAKAL dataset\n");
            if (!in_line) fclose(impin);
            return;
        default:
            break;    /* end of default */
    }   /* end of switch */
  }
}  /* end of import_schakal */

